#!/usr/bin/env python

import random
import urllib2 
from socket import gethostbyname
from threading import Thread, Timer
import getopt
import sys
import signal
import re
import time
from PassiveFishTPClient import PassiveFishTPClient
import FishTPServer
from blogfish_functions import debug


def _tidy(addr):
		try:
			(host, port) = addr.split(':', 2)
			host = gethostbyname(host)
			return host + ":" + port
		except:
			return False # hosts are never false, so this is ok

def _is_passive(addr):
	return addr.endswith(":PASV")
	
class PeerManager:

	def __init__(self,plf, mp, mf, port, ffi):
		self.peers = []
		self.asked_peers = []
		self.poss_peers = []
		self.fishes = []
		self.fishtp_server = FishTPServer.FishTPServer(port, self)
		self.failures = {}
		self.socketmap = {}
		
		self.peerListFrom = plf
		self.maxPeers = mp
		self.maxFish = mf
		self.port = port
		self.flush_fish_interval = ffi

		for sig in (signal.SIGINT, signal.SIGQUIT, signal.SIGHUP):
			signal.signal(sig, signal.SIG_IGN)

		signal.signal(signal.SIGTERM, self._handle_signal)

	
	def run(self):
		self.timeout_add(30, self._manage_peers)
		self.timeout_add(300, self._look_for_peers)
		self.timeout_add(self.flush_fish_interval, self._flush_fish)
		self.fishtp_server.start()
		self._look_for_peers()
		while True: time.sleep(60)


	def timeout_add(self, secs, function):
		l = lambda: self.timeout_add(secs, function) and function() 
		t = Timer(secs, l)
		t.setDaemon(True)
		t.start()
		return True

	def handle_message(self, string, remote_host, wfile=None):
		debug("handling message %s" % string)
		peerAddr = remote_host + ":" + self._get_port_from_message(string)
		peerAddr = _tidy(peerAddr)
		if not peerAddr: 
			debug("Got bad peer address")
			return
		if string.startswith("JOIN PASV"):
			self._consider_peer(peerAddr, wfile) # never returns
		elif string.startswith("JOIN"):
			self._consider_peer(peerAddr)
		elif string.startswith("ACK"):
			self._handle_peer_ack(peerAddr)
		elif string.startswith("QUIT"):
			self._remove_peer(peerAddr)
		# those below REQUIRE a peer connection
		elif peerAddr not in self.peers:
			debug("Got message from a peer I don't know... ")
			# assume they think we know them
			if _is_passive(peerAddr): self._consider_peer(peerAddr, wfile) 
			else: self._consider_peer(peerAddr)
		elif string.startswith("NBRS?"):
			self._handle_peer_neighbour_request(peerAddr)
		elif string.startswith("NBRS"):
			poss_peers = string.splitlines()
			poss_peers.pop(0)
			self._handle_peer_neighbour_response(peerAddr, poss_peers)
		elif string.startswith("FISH"):
			fstr = "".join(string.splitlines(True)[1:]) # remove first line
			if peerAddr in self.peers:
				debug("adding fish from neighbour")
				self.fishes.append(fstr)				
			else:
				debug("Got fish from unknown address '%s'" % peerAddr)
		elif self._has_peer_at(remote_host): # compatibility with old version
			self.fishes.append(string)
		else:
			debug("got fish from unknown host '%s'" % remote_host)

	def _get_port_from_message(self, string):
		match = re.match('\S+\s+(\S+)\n', string)
		try:
			return match.group(1)
		except:
			debug("Couldn't get port from message '%s'" % string)
	
	def _manage_peers(self):
		"""called every few seconds, in order to manage list of peers"""
		debug("got %s peers" % len(self.peers))
		debug("got %s fish" % len(self.fishes))
		debug("time is " + time.asctime())
		
		for peer in self.peers:
			debug("Peer: %s" % peer)
		if len(self.peers) == self.maxPeers: return True
		while len(self.peers) > self.maxPeers:
			self.remove_peer(random.choice(self.peers))
		
		for i in range(1,3): 
			if self.poss_peers:
				peerAddr = self.poss_peers.pop(0)
				peerAddr = peerAddr.strip() # remove newlines and junk
				self.try_peer(peerAddr)
			if len(self.peers) > self.maxPeers: break
	

	def _flush_fish(self):
		debug("in _flush_fish")
		if not self.fishes: return
		if not self.peers: return
		peer = random.choice(self.peers)
		while True: # always send away at least one fish
			fstr = self.fishes.pop()
			self.send_fish_away(fstr)
			if len(self.fishes) <= self.maxFish: break


	def _look_for_peers(self):
		"""called less commonly, more network intense"""
		debug("_look_for_peers")
		if len(self.peers) >= self.maxPeers: return True
		
		if len(self.peerListFrom):
			try:
				debug("trying peerList " + self.peerListFrom)	
				poss_peers = urllib2.urlopen(self.peerListFrom, \
			   		"port=" + self.port).readlines()
				 
				for peer in poss_peers:
					if peer not in self.poss_peers:
						self.poss_peers.append(peer)
			except:
				debug("Couldn't look up " + self.peerListFrom + " to get peers")
		
		if len(self.peers):
			debug("sending NBRS request")
			self.send_to_random("NBRS? %s\n" % self.port)
	

	def get_peers(self):
		return self.peers[:]
	
	def get_asked_peers(self):
		return self.asked_peers[:]
		
	def _has_peer_at(self, host):
		host = gethostbyname(host)
		for peer in self.peers:
			(phost, port) = peer.split(":", 2)
			if phost == host: return True
		return False
		
	def _consider_peer(self, addr, wfile=None):
		"""Consider a peer for adding, in response to a request"""
		if len(self.peers) >= self.maxPeers: return
		addr = _tidy(addr)
		if addr == gethostbyname("localhost") + ":%s" % self.port: return
		if not addr: return

		self._add_peer(addr, wfile)
		self.send_message(addr, "ACK %s\n" % self.port)
		if wfile:				
			self.socketmap[addr].wait_forever()	
			# NEVER RETURNS!!! This prevents socket closure 
			# in the ThreadingHTTPServer
		
	def try_peer(self, addr):
		"""Send an 'add me' request to a peer"""
		if len(self.peers) >= self.maxPeers: return
		addr = _tidy(addr)
		debug("trying peer at %s" % addr)
		if addr and addr not in self.peers and addr not in self.asked_peers:
			self.asked_peers.append(addr)
			if self.port == "PASV":
				# we need to keep the connection open, so don't
				# use a separate thread.
				if _is_passive(addr):
					debug("We tried to link to another passive peer - " \
							+ "this is not possible")
					return
				self.fishtp_server.attempt_join(addr)
			else:
				self.send_message(addr, "JOIN %s\n" % self.port)

	def _add_peer(self, addr, wfile=None):
		if addr not in self.peers: 
			self.peers.append(addr)
		if addr in self.asked_peers: self.asked_peers.remove(addr)
		if addr in self.poss_peers: self.poss_peers.remove(addr)
		# inform FishTPServer so passive servers can add a connection
		# this should be ignored if we are already joined
		self.fishtp_server.attempt_join(addr)
		if wfile:
			# we must be active, or how did we get this message?
			self.socketmap[addr] = PassiveFishTPClient(wfile, addr, self)
			

	def _remove_peer(self, addr):
		"""remove a peer, without sending a message"""
		if addr in self.peers: 
			self.peers.remove(addr)
		if addr in self.asked_peers: 
			self.asked_peers.remove(addr)
		if addr in self.poss_peers: self.poss_peers.remove(addr)
		if self.socketmap.has_key(addr): del self.socketmap[addr]
	
	def remove_peer(self, addr):
		"""remove a peer from my list"""
		addr = _tidy(addr)
		if addr:
			self.send_message(addr, "QUIT %s\n" % self.port)
			self._remove_peer(addr)
	
	def send_fish_away(self, fstr):
		debug("sending fish away")
		fstr = "FISH %s\n" % self.port + fstr
		self.send_to_random(fstr)
	
	def _handle_peer_ack(self, addr):
		"""handle an ack message from another peer"""
		addr = _tidy(addr)
		debug("got ack from " + addr)
		if addr and addr in self.asked_peers: 
			self._add_peer(addr)

	def _handle_peer_neighbour_request(self, addr):
		"""handle a request for some more peer addresses"""
		addr = _tidy(addr)
		debug("got nbr req from " + addr)
		if addr and addr in self.peers:
			peers = self.peers[:]
			peers.remove(addr)
			if not len(peers): return
			chosen = []
			msg = "NBRS %s\n" % self.port
			if len(peers) > 1:
				while len(chosen) < min(3, len(peers)):
					p = random.choice(peers)
					if p not in chosen: chosen += p
				# Python 2.3 only...
				# chosen = random.sample(peers, min(3, len(peers)))
				for p in chosen: msg = msg + p + "\n"
			else:
				msg = msg + peers[0] + "\n"
			self.send_message(addr, msg)
			
	def _handle_peer_neighbour_response(self, addr, poss_addrs):
		addr = _tidy(addr)
		debug("got nbr resp from " + addr)
		if addr and addr in self.peers:
			for poss in poss_addrs:
				poss = _tidy(poss)
				if poss and not _is_passive(poss): 
					self.poss_peers.append(poss)
	
	def send_to_random(self, message):
		"""Send a message to a random peer"""
		if not self.peers: return
		peer = random.choice(self.peers)
		self.send_message(peer, message)
	
	def send_message(self, addr, message):
		"""Send a message to a peer"""
		addr = _tidy(addr)
		if _is_passive(addr):
			t = Thread(None, self._passive_send, None, (addr, message))
		else:
			t = Thread(None, self._tryurl, None, (addr, message))
		t.setDaemon(True)
		t.start()
		
	def shutdown(self):
		for peer in self.peers:
			self.remove_peer(peer)
		for fish in self.fishes:
			self.send_fish_away(fish)
		i = 10
		while i:
			print "%s... " % i
			i -= 1
			time.sleep(1)
		print "Exiting"
		sys.exit(0)
	
	
	def _tryurl(self, addr, data):
		try:
			debug("sending:" + data[0:20] + " ...\nto " + addr)
			urllib2.urlopen("http://" + addr, data)
			self.handle_success(addr)
		except: # avoid error messages
			debug("failed to send to " + addr)
			self.handle_failure(addr)
			pass

	def _passive_send(self, addr, data):
		if addr in self.socketmap:
			self.socketmap[addr].push(data) # automatically tries to write
		else:
			debug("No socket available for %s" % addr)
			self.handle_failure(addr)
	
	def handle_failure(self, addr):
		debug("handling failure for '%s'" % addr)
		if addr in self.asked_peers: self._remove_peer(addr) # no tolerance
		if addr in self.socketmap: del self.socketmap[addr] # kill thread?
		if addr in self.peers:
			if addr in self.failures:	self.failures[addr] +=1
			else: self.failures[addr] = 1
			if self.failures[addr] > 2:
				debug("removing peer '%s' as 3 failures noted" % addr)
				del self.failures[addr]
				# further failures will send no msg, so recursion is ok
				self.remove_peer(addr) 

	def handle_success(self, addr):
		# reset the count
		self.failures[addr] = 0
		
	def _handle_signal(self, signum, frame):
		debug("Got signal, quitting")
		self.shutdown()

if __name__ == "__main__":
	try:
		opts, args = getopt.getopt(sys.argv[1:], "df:p:m:u:i:")
		opts = dict(opts)
		for k in "iupmf": 
			k = "-%s" % k
			if k not in opts:
				opts[k] = 0

		plf = opts["-u"] or "http://blogfish.sourceforge.net/cgi-bin/peerlist.cgi"
		port = opts["-p"] or "5156"
		port = str(port)
		if "-d" in opts.keys(): 
			import blogfish_functions
			blogfish_functions.Debug = True
			debug("debug mode on")
		debug("serving on port %s" % port)
		mp = int(opts["-m"]) or 5
		mf = int(opts["-f"]) or 10
		ffi = int(opts["-i"]) or 600
	except:
		# raise
		print """

Usage:
  PeerManager.py -p <port> -u <url for other peers> -f <maximum no of fish>
  -m <maximum no of peers> -i <seconds interval before sending fish away> [-d]
 
  Use -d to debug

"""
		sys.exit(1)

	debug("starting up at " + time.asctime())
	PeerManager(plf, mp, mf, port, ffi).run()
